/*
 * $Id: RADIUSauth.cpp,v 1.18 2005/03/03 18:04:23 hweiss Exp $
 *
 * RADIUS Plugin for pGina
 * Copyright (c) 2004, 2005 Holger Weiss <holger@ZEDAT.FU-Berlin.DE>
 * Copyright (c) 2005 Kostas Kalevras <kkalev@noc.ntua.gr>
 *
 * ---------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * ---------------------------------------------------------------------------
 *
 * This code is based on pam_radius_auth, copyright (c) 2003 Alan DeKok.
 */

#include "stdafx.h"
#include "RADIUSauth.h"
#include "RADIUShelpers.h"
#include "radius.h"
#include "md5.h"

/* save statically for all functions */
static bool     debug = false;
static bool     userlogon = false;
#define DPRINT if (debug) radiusLog

/* save statically between requests */
static request_t modulePrivate;

/* ---------------------------------------------------------------------------
 * Core RADIUS code
 */

/*
 * Convert server->hostname to server->ip and set server->port
 */
static int
host2server(radius_server_t *server)
{
	static bool     accounting_start = true;

	if ((!modulePrivate.server) && ((server->ip.sin_addr.s_addr = get_ipaddr(server->hostname)) == ((UINT4) 0))) {
		DPRINT(EVENTLOG_ERROR_TYPE, "get_ipaddr(%s) returned 0", server->hostname);
		return RADIUS_AUTHINFO_UNAVAIL;
	}
	/*
	 * If server->port is set, increment it by 1 for accounting start requests.
	 * Yes, this is very ugly.  We should simplify things by adding a seperate
	 * port for accounting to radius_server_t.
	 */
	if (server->port) {
		if (server->accounting) {
			if (accounting_start) {
				server->port = htons(ntohs(server->port) + 1);
				accounting_start = false;
			} else {
				accounting_start = true;
			}
		}
	/* if the server port hasn't already been defined, go get it */
	} else {
		struct servent *svp;

		if (!server->accounting)
			svp = getservbyname("radius", "udp");
		else
			svp = getservbyname("radacct", "udp");

		if (svp == (struct servent *) 0) {
			DPRINT(EVENTLOG_ERROR_TYPE, "getservbyname failed to lookup radius or radacct port");
			return RADIUS_AUTHINFO_UNAVAIL;
		}
		server->port = svp->s_port;
	}

	return RADIUS_SUCCESS;
}

/*
 * Get a pseudo-random vector
 */
static void
get_random_vector(unsigned char *vector)
{
	MD5_CTX         my_md5;
	struct timeval  tv;
	struct timezone tz;
	static unsigned int session = 0;	/* make the number harder to guess */
	unsigned long rnd;

	/*
	 * Use the time of day with the best resolution the system can give
	 * us -- often close to microsecond accuracy
	 */
	gettimeofday(&tv, &tz);

	if (session == 0)
		session = getpid();

	sgenrand(tv.tv_sec + (unsigned long)session++);
	rnd = genrand();

	/*
	 * Hash things to get maybe cryptographically strong pseudo-random
	 * numbers
	 */
	MD5Init(&my_md5);
	MD5Update(&my_md5, (unsigned char *)&tv, sizeof(tv));
	MD5Update(&my_md5, (unsigned char *)&tz, sizeof(tz));
	MD5Update(&my_md5, (unsigned char *)&rnd, sizeof(rnd));
	MD5Final(vector, &my_md5);	/* set the final vector */
}

/*
 * RFC 2139 says to generate the accounting request vector this way. However,
 * the Livingston 1.16 server doesn't check it.  The Cistron server
 * (http://www.radius.cistron.nl/) does, and this code seems to work with it.
 * It also works with Funk's Steel-Belted RADIUS.
 */
static void
get_accounting_vector(AUTH_HDR *request, radius_server_t *server)
{
	MD5_CTX         my_md5;
	int             secretlen = (int)strlen(server->secret);
	int             len = ntohs(request->length);

	memset(request->vector, 0, AUTH_VECTOR_LEN);
	MD5Init(&my_md5);
	memcpy(((char *)request) + len, server->secret, secretlen);

	MD5Update(&my_md5, (unsigned char *)request, len + secretlen);
	MD5Final(request->vector, &my_md5);	/* set the final vector */
}

/*
 * Verify the response from the server
 */
static int
verify_packet(char *secret, AUTH_HDR *response, AUTH_HDR *request)
{
	MD5_CTX         my_md5;
	unsigned char   calculated[AUTH_VECTOR_LEN];
	unsigned char   reply[AUTH_VECTOR_LEN];

	/*
	 * We could dispense with the memcpy, and do MD5's of the packet +
	 * vector piece by piece.  This is easier to understand, and maybe
	 * faster.
	 */
	memcpy(reply, response->vector, AUTH_VECTOR_LEN);	/* save the reply */
	memcpy(response->vector, request->vector, AUTH_VECTOR_LEN);	/* sent vector */

	/* MD5(response packet header + vector + response packet data + secret) */
	MD5Init(&my_md5);
	MD5Update(&my_md5, (unsigned char *)response, ntohs(response->length));

	/*
	 * This next bit is necessary because of a bug in the original
	 * Livingston RADIUS server.  The authentication vector is *supposed*
	 * to be MD5'd with the old password (as the secret) for password
	 * changes.  However, the old password isn't used.  The
	 * "authentication" vector for the server reply packet is simply the
	 * MD5 of the reply packet.  Odd, the code is 99% there, but the old
	 * password is never copied to the secret!
	 */
	if (*secret)
		MD5Update(&my_md5, (unsigned char *)secret, (unsigned int)strlen(secret));

	MD5Final(calculated, &my_md5);	/* set the final vector */

	/* did he use the same random vector + shared secret? */
	if (memcmp(calculated, reply, AUTH_VECTOR_LEN) != 0)
		return FALSE;

	return TRUE;
}

/*
 * Find an attribute in a RADIUS packet.  Note that the packet length is
 * *always* kept in network byte order.
 */
static attribute_t *
find_attribute(AUTH_HDR *response, unsigned char type)
{
	attribute_t    *attr = (attribute_t *) & response->data;
	int             len = ntohs(response->length) - AUTH_HDR_LEN;

	while (attr->attribute != type) {
		if ((len -= attr->length) <= 0)
			return NULL;	/* not found */
		attr = (attribute_t *) ((char *)attr + attr->length);
	}

	return attr;
}

/*
 * Find a vendor specific attribute in a RADIUS packet
 */
static vendor_attr_t *
find_vendor_attribute(AUTH_HDR *response, unsigned int vendor, unsigned char type)
{
	attribute_t	*attr = (attribute_t *) & response->data;
	vendor_attr_t	*vndr;
	int		len = ntohs(response->length) - AUTH_HDR_LEN;
	unsigned int	i = 0;

	while(1){
		while (attr->attribute != PW_VENDOR_SPECIFIC) {
			if ((len -= attr->length) <= 0)
				return NULL;	/* not found */
			if (attr->length <= 2)	/* Too small attribute */
				return NULL;
			attr = (attribute_t *) ((char *)attr + attr->length);
		}
		if (attr->length <= 6)	/* Too small attribute for vendor specific */
			return NULL;
		vndr = (vendor_attr_t *) ((char *)attr + 2);
		if (vndr->length <= 2)	/* Too small attribute */
			return NULL;
		if (ntohl(vndr->vendor) == vendor && vndr->attribute == type){
			return vndr;
		}else{
			if ((len -= attr->length) <= 0)
				return NULL;	/* not found */
			attr = (attribute_t *) ((char *)attr + attr->length);
		}
	}

	return NULL;
}

/*
 * Changes an attribute value. Length MUST be the same. Returns 1 (true) on success
 */
static bool
change_attribute(attribute_t *attr, const unsigned char *data, int length)
{
	if (attr != NULL && data != NULL && attr->attribute != PW_VENDOR_SPECIFIC){
		if (attr->length == (unsigned char)length + 2){
			memcpy(attr->data, data, length);
			return true;
		}
	}

	return false;
}

/*
 * Changes an integer attribute value. Returns 1 (true) on success
 */
static bool
change_int_attribute(attribute_t *attr, int data)
{
	int	value = htonl(data);

	return change_attribute(attr, (unsigned char *)&value, sizeof(int));
}

/*
 * Changes a vendor attribute value. Length MUST be the same. Returns 1 (true) on success
 */
static bool
change_vendor_attribute(vendor_attr_t *attr, const unsigned char *data, int length)
{
	if (attr != NULL && data != NULL){
		if (attr->length == (unsigned char) length + 2){
			memcpy(attr->data, data, length);
			return true;
		}
	}

	return false;
}

/*
 * Changes an integer vendor attribute value. Returns 1 (true) on success
 */
static bool
change_vendor_int_attribute(vendor_attr_t *attr, int data)
{
	int	value = htonl(data);

	return change_vendor_attribute(attr, (unsigned char *)&value, sizeof(int));
}

/*
 * Add an attribute to a RADIUS packet
 */
static void
add_attribute(AUTH_HDR *request, unsigned char type, const unsigned char *data, int length)
{
	attribute_t    *p;

	if (request == NULL || type == PW_VENDOR_SPECIFIC
	    || data == NULL || length > AUTH_STRING_LEN)
		return;

	if (ntohs(request->length) + length + 2 > RADIUS_PACKET_LEN)
		return;

	p = (attribute_t *) ((unsigned char *)request + ntohs(request->length));
	p->attribute = type;
	p->length = (unsigned char)length + 2;	/* the total size of the attribute */
	request->length = htons(ntohs(request->length) + p->length);
	memcpy(p->data, data, length);
}

/*
 * Add a vendor attribute to the RADIUS packet
 */
static void
add_vendor_attribute(AUTH_HDR *request, unsigned int vendor, unsigned char type,
                     const unsigned char *data, int length)
{
	attribute_t	*q;
	vendor_attr_t	*p;

	if (request == NULL || data == NULL || length > AUTH_STRING_LEN - 6)
		return;

	if (ntohs(request->length) + length + 8 > RADIUS_PACKET_LEN)
		return;

	q = (attribute_t *) ((unsigned char *)request + ntohs(request->length));
	q->attribute = PW_VENDOR_SPECIFIC;
	q->length = (unsigned char) length + 8;
	p = (vendor_attr_t *) ((unsigned char *)request + ntohs(request->length) + 2);
	p->vendor = htonl(vendor);
	p->attribute = type;
	p->length = (unsigned char) length + 2;
	request->length = htons(ntohs(request->length) + q->length);
	memcpy(p->data, data, length);
}

/*
 * Add an integer vendor attribute to a RADIUS packet
 */
static void
add_vendor_int_attribute(AUTH_HDR *request, unsigned int vendor, unsigned char type, int data)
{
	int	value = htonl(data);

	add_vendor_attribute(request, vendor, type, (unsigned char *)&value, sizeof(int));
}

/*
 * Add an integer attribute to a RADIUS packet
 */
static void
add_int_attribute(AUTH_HDR *request, unsigned char type, int data)
{
	int             value = htonl(data);

	add_attribute(request, type, (unsigned char *)&value, sizeof(int));
}

/*
 * Get the attribute value in string format. Do some checks on the attribute first
 * Returns NULL on failure
 */
static char
*get_string_attribute_value(attribute_t *p)
{
	if (p != NULL && p->attribute != PW_VENDOR_SPECIFIC && p->length > 2){
		static char value[AUTH_STRING_LEN + 1];

		memcpy(value,p->data, p->length - 2);
		value[p->length - 2] = '\0';

		return value;
	}

	return NULL;
}

/*
 * Get the attribute integer value. Do some checks on the attribute first.
 * Returns false on failure
 */
static bool
get_int_attribute_value(attribute_t *p, unsigned int *value)
{
	if (p != NULL && p->attribute != PW_VENDOR_SPECIFIC && p->length == 6 && value != NULL){
		memcpy(value,p->data, p->length - 2);
		return true;
	}

	return false;
}

/*
 * Get the vendor attribute value in string format. Do some checks on the attribute first
 * Returns NULL on failure
 */
static char
*get_vendor_string_attribute_value(vendor_attr_t *p)
{
	if (p != NULL && p->length > 2){
		static char value[AUTH_STRING_LEN + 1];

		memcpy(value,p->data, p->length - 2);
		value[p->length - 2] = '\0';

		return value;
	}

	return NULL;
}

/*
 * Get the vendor attribute integer value. Do some checks on the attribute first.
 * Returns false on failure
 */
static bool
get_vendor_int_attribute(vendor_attr_t *p, unsigned int *value)
{
	if (p != NULL && p->length == 6){
		memcpy(value,p->data, p->length - 2);
		return true;
	}

	return false;
}

/*
 * Add a RADIUS password attribute to the packet.  Some magic is done here.
 * If it's an PW_OLD_PASSWORD attribute, it's encrypted using the encrypted
 * PW_PASSWORD attribute as the initialization vector.  If the password
 * attribute already exists, it's over-written.  This allows us to simply call
 * add_password to update the password for different servers.
 */
static void
add_password(AUTH_HDR *request, unsigned char type, const char *password, char *secret)
{
	MD5_CTX         md5_secret, my_md5;
	unsigned char   misc[AUTH_VECTOR_LEN];
	int             i;
	int             length = (int)strlen(password);
	unsigned char   hashed[256 + AUTH_PASS_LEN];	/* can't be longer than this */
	unsigned char  *vector;
	attribute_t    *attr;

	if (length > MAXPASS)	/* shorten the password for now */
		length = MAXPASS;
	if (length == 0)
		length = AUTH_PASS_LEN;	/* 0 maps to 16 */
	if ((length & (AUTH_PASS_LEN - 1)) != 0) {
		length += (AUTH_PASS_LEN - 1);	/* round it up */
		length &= ~(AUTH_PASS_LEN - 1);	/* chop it off */
	}			/* 16*N maps to itself */
	memset(hashed, 0, length);
	memcpy(hashed, password, strlen(password));

	attr = find_attribute(request, PW_PASSWORD);

	if (type == PW_PASSWORD)
		vector = request->vector;
	else
		vector = attr->data;	/* attr CANNOT be NULL here */

	/*
	 * Encrypt the password:
	 * password : e[0] = p[0] ^ MD5(secret + vector)
	 */
	MD5Init(&md5_secret);
	MD5Update(&md5_secret, (unsigned char *)secret, (unsigned int)strlen(secret));
	my_md5 = md5_secret;	/* so we won't re-do the hash later */
	MD5Update(&my_md5, vector, AUTH_VECTOR_LEN);
	MD5Final(misc, &my_md5);	/* set the final vector */
	xor(hashed, misc, AUTH_PASS_LEN);

	/* for each step through, e[i] = p[i] ^ MD5(secret + e[i-1]) */
	for (i = 1; i < (length >> 4); i++) {
		my_md5 = md5_secret;	/* grab old value of the hash */
		MD5Update(&my_md5, &hashed[(i - 1) * AUTH_PASS_LEN], AUTH_PASS_LEN);
		MD5Final(misc, &my_md5);	/* set the final vector */
		xor(&hashed[i * AUTH_PASS_LEN], misc, AUTH_PASS_LEN);
	}

	if (type == PW_OLD_PASSWORD)
		attr = find_attribute(request, PW_OLD_PASSWORD);
	if (!attr)
		add_attribute(request, type, hashed, length);
	else
		memcpy(attr->data, hashed, length);	/* overwrite the packet */
}

/*
 * Return pointer to duplicated radius_server_t *server
 */
static radius_server_t *
radius_dup(radius_server_t *server)
{
	radius_server_t *new_server = (radius_server_t *) malloc(sizeof(radius_server_t));

	memcpy(new_server, server, sizeof(radius_server_t));
	new_server->hostname = strdup(server->hostname);
	new_server->secret = strdup(server->secret);

	return new_server;
}

/*
 * Free server->hostname, server->secret and server
 */
static void
cleanup(radius_server_t *server)
{
	radius_server_t *next;

	while (server) {
		next = server->next;
		_radius_drop(server->hostname);
		_radius_forget(server->secret);
		_radius_drop(server);
		server = next;
	}
}

/*
 * Allocate and open a local port for communication with the RADIUS server
 */
static int
initialize(radius_conf_t *conf, int accounting)
{
	struct sockaddr salocal;
	int             timeout;
	u_short         local_port;
	char           *hostname = NULL;
	radius_server_t *tmp, *server = NULL;
	struct sockaddr_in *s_in;

	timeout = regReadUlong("timeout") ? (int)regReadUlong("timeout") : 3;

	if (modulePrivate.server) {
		conf->server = radius_dup(modulePrivate.server);
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Reusing server %s", conf->server->hostname);
	} else {
		/* server 1 */
		if ((hostname = regReadString("server1")) != NULL) {
			/* pointer returned by malloc must be casted in C++ */
			tmp = (radius_server_t *) malloc(sizeof(radius_server_t));
			conf->server = tmp;
			server = tmp;
			if ((server->secret = regReadString("secret1")) == NULL) {
				DPRINT(EVENTLOG_ERROR_TYPE, "No shared secret given for server %s", hostname);
				_radius_drop(hostname);
				_radius_drop(server);
				return RADIUS_AUTHINFO_UNAVAIL;
			}
			server->hostname = hostname;
			server->port = htons((u_short)regReadUlong("port1"));	/* returns 0 if not specified */
			server->accounting = accounting;
			server->timeout = timeout;
			server->next = NULL;
			DPRINT(EVENTLOG_INFORMATION_TYPE, "Server %s is configured", server->hostname);
		}

		/* server 2 */
		if ((hostname = regReadString("server2")) != NULL) {
			tmp = (radius_server_t *) malloc(sizeof(radius_server_t));
			if (server)
				server->next = tmp;
			else
				conf->server = tmp;
			server = tmp;
			if ((server->secret = regReadString("secret2")) == NULL) {
				DPRINT(EVENTLOG_ERROR_TYPE, "No shared secret given for server %s", hostname);
				_radius_drop(hostname);
				_radius_drop(server);
				return RADIUS_AUTHINFO_UNAVAIL;
			}
			server->hostname = hostname;
			server->port = htons((u_short)regReadUlong("port2"));
			server->accounting = accounting;
			server->timeout = timeout;
			server->next = NULL;
			DPRINT(EVENTLOG_INFORMATION_TYPE, "Server %s is configured", server->hostname);
		}

		/* server 3 */
		if ((hostname = regReadString("server3")) != NULL) {
			tmp = (radius_server_t *) malloc(sizeof(radius_server_t));
			if (server)
				server->next = tmp;
			else
				conf->server = tmp;
			server = tmp;
			if ((server->secret = regReadString("secret3")) == NULL) {
				DPRINT(EVENTLOG_ERROR_TYPE, "No shared secret given for server %s", hostname);
				_radius_drop(hostname);
				_radius_drop(server);
				return RADIUS_AUTHINFO_UNAVAIL;
			}
			server->hostname = hostname;
			server->port = htons((u_short)regReadUlong("port3"));
			server->accounting = accounting;
			server->timeout = timeout;
			server->next = NULL;
			DPRINT(EVENTLOG_INFORMATION_TYPE, "Server %s is configured", server->hostname);
		}

		if (!server) {
			/* no server found, die a horrible death */
			DPRINT(EVENTLOG_ERROR_TYPE, "No RADIUS server configured");
			return RADIUS_AUTHINFO_UNAVAIL;
		}
	}

	/* open a socket (dies if it fails) */
	conf->sockfd = (int)socket(AF_INET, SOCK_DGRAM, 0);
	if (conf->sockfd < 0) {
		DPRINT(EVENTLOG_ERROR_TYPE, "Failed to open RADIUS socket: %s", strerror(errno));
		return RADIUS_AUTHINFO_UNAVAIL;
	}

	/* set up the local end of the socket communications */
	s_in = (struct sockaddr_in *) & salocal;
	memset((char *)s_in, '\0', sizeof(struct sockaddr));
	s_in->sin_family = AF_INET;
	s_in->sin_addr.s_addr = INADDR_ANY;

	if (local_port = (u_short)regReadUlong("local_port")) {
		/* use the local port specified by the user */
		s_in->sin_port = htons(local_port);
		if (bind(conf->sockfd, &salocal, sizeof(struct sockaddr_in))) {
			closesocket(conf->sockfd);
			DPRINT(EVENTLOG_ERROR_TYPE, "Can't bind to port %d", local_port);
			return RADIUS_AUTHINFO_UNAVAIL;
		}
	} else {
		/* use our process ID as a local port for RADIUS */
		local_port = ((unsigned short)getpid() & 0x7fff) + 1024;
		do {
			local_port++;
			s_in->sin_port = htons(local_port);
		} while ((bind(conf->sockfd, &salocal, sizeof(struct sockaddr_in)) < 0)
		         && (local_port < 64000));

		if (local_port >= 64000) {
			closesocket(conf->sockfd);
			DPRINT(EVENTLOG_ERROR_TYPE, "No open port we could bind to");
			return RADIUS_AUTHINFO_UNAVAIL;
		}
	}
	DPRINT(EVENTLOG_INFORMATION_TYPE, "Using local port %d", local_port);

	return RADIUS_SUCCESS;
}

/*
 * Helper function for building a radius packet.  It initializes *some* of the
 * header, and adds common attributes.
 */
static void
build_radius_packet(AUTH_HDR *request, const char *user, const char *password, radius_conf_t *conf)
{
	char           *hostname;
	const char     *rhost;
	UINT4           ipaddr;

	hostname = regReadString("nas");
	if (hostname == NULL) {
		hostname = (char *)malloc(256);
		*hostname = '\0';
		gethostname(hostname, 256);
	}

	request->length = htons(AUTH_HDR_LEN);

	if (password)	/* make a random authentication req vector */
		get_random_vector(request->vector);

	add_attribute(request, PW_USER_NAME, (unsigned char *)user, (int)strlen(user));

	/* add a password, if given */
	if (password)
		add_password(request, PW_PASSWORD, password, conf->server->secret);

	/* add a NULL password to non-accounting requests */
	else if (request->code != PW_ACCOUNTING_REQUEST)
		add_password(request, PW_PASSWORD, "", conf->server->secret);

	/* the packet is from localhost if on localhost, to make configs easier */
	if ((conf->server->ip.sin_addr.s_addr == ntohl(0x7f000001)) || (!*hostname))
		ipaddr = 0x7f000001;
	else {
		struct hostent *hp;

		if ((hp = gethostbyname(hostname)) == (struct hostent *) NULL) {
			ipaddr = 0x00000000;	/* no client IP address */
		} else {
			ipaddr = ntohl(*(UINT4 *) hp->h_addr);	/* use the first one available */
		}
	}

	/* add either IP address or host name */
	if (ipaddr) {
		add_int_attribute(request, PW_NAS_IP_ADDRESS, ipaddr);
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Using NAS IP address attribute");
	} else {
		add_attribute(request, PW_NAS_IDENTIFIER, (unsigned char *)&hostname, (int)strlen(hostname));
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Using NAS identifier attribute (%s)", hostname);
	}

	/* add in the port type (virtual) */
	add_int_attribute(request, PW_NAS_PORT_TYPE, PW_NAS_PORT_TYPE_VIRTUAL);
	
	/* add in a random port number based on the pid */
	add_int_attribute(request, PW_NAS_PORT_ID, getpid());

	/* tell the server which host the user is coming from */
	if ((rhost = getHostID()) != NULL) {
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Got host IP %s", rhost);
		add_attribute(request, PW_CALLING_STATION_ID, (unsigned char *)rhost,
		              (int)strlen(rhost));
	}

	/* If we are also supporting userlogon attributes send the userlogon-type */
	if (userlogon)
		add_vendor_int_attribute(request, PW_VENDOR_NTUA, PW_USERLOGON_TYPE, PW_USERLOGON_TYPE_WINDOWS_LOGON);

	free(hostname);
}

/*
 * Talk RADIUS to a server.  Send a packet and get the response.
 */
static int
talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response,
            char *password, char *old_password, int tries)
{
	int             salen, total_length;
	fd_set          set;
	struct timeval  tv;
	time_t          now, end;
	int             rcode;
	struct sockaddr saremote;
	struct sockaddr_in *s_in = (struct sockaddr_in *) & saremote;
	radius_server_t *server = conf->server;
	int             livingston_bug = regReadBool("livingston");
	int             ok;
	int             server_tries;
	int             retval;

	/* loop over all available servers */
	while (server != NULL) {

		/* only look up IP information as necessary */
		if ((retval = host2server(server)) != RADIUS_SUCCESS) {

			DPRINT(EVENTLOG_ERROR_TYPE,
			       "Failed looking up IP address for RADIUS server %s (errcode=%d)",
			       server->hostname, retval);
			ok = FALSE;
			goto next;	/* skip to the next server */
		}

		/* set up per-server IP && port configuration */
		memset((char *)s_in, '\0', sizeof(struct sockaddr));
		s_in->sin_family = AF_INET;
		s_in->sin_addr.s_addr = htonl(server->ip.sin_addr.s_addr);
		s_in->sin_port = server->port;
		total_length = ntohs(request->length);

		if (!password)	/* make an RFC 2139 p6 request authenticator */
			get_accounting_vector(request, server);

		server_tries = tries;
send:
		/* send the packet */
		if (sendto(conf->sockfd, (char *)request, total_length, 0,
		           &saremote, sizeof(struct sockaddr_in)) < 0) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Error sending RADIUS packet to server %s: %s",
			       server->hostname, strerror(errno));
			ok = FALSE;
			goto next;	/* skip to the next server */
		}

		/* wait for the response, and verify it */
		salen = sizeof(struct sockaddr);
		tv.tv_sec = server->timeout;	/* wait for the specified time */
		tv.tv_usec = 0;
		FD_ZERO(&set);	/* clear out the set */
		FD_SET(conf->sockfd, &set);	/* wait only for the RADIUS UDP socket */

		time(&now);
		end = now + tv.tv_sec;

		/* loop, waiting for the select to return data */
		ok = TRUE;
		while (ok) {

			rcode = select(conf->sockfd + 1, &set, NULL, NULL, &tv);

			/* select timed out */
			if (rcode == 0) {
				DPRINT(EVENTLOG_ERROR_TYPE, "RADIUS server %s failed to respond",
				       server->hostname);
				if (--server_tries)
					goto send;
				ok = FALSE;
				break;	/* exit from the select loop */
			} else if (rcode < 0) {

				/* select had an error */
				if (errno == EINTR) {	/* we were interrupted */
					time(&now);

					if (now > end) {
						DPRINT(EVENTLOG_ERROR_TYPE, "RADIUS server %s failed to respond",
						       server->hostname);
						if (--server_tries)
							goto send;
						ok = FALSE;
						break;	/* exit from the select loop */
					}
					tv.tv_sec = (long)end - (long)now;
					if (tv.tv_sec == 0) {	/* keep waiting */
						tv.tv_sec = 1;
					}
				} else {	/* not an interrupt, it was a real error */
					DPRINT(EVENTLOG_ERROR_TYPE, "Error waiting for response from RADIUS server %s: %s",
					       server->hostname,
					       strerror(errno));
					ok = FALSE;
					break;
				}

			/* the select returned OK */
			} else if (FD_ISSET(conf->sockfd, &set)) {

				/* try to receive some data */
				if ((total_length = recvfrom(conf->sockfd, (char *)response,
				                             RADIUS_PACKET_LEN, 0, &saremote, &salen)) < 0) {
					DPRINT(EVENTLOG_ERROR_TYPE, "error reading RADIUS packet from server %s: %s",
					       server->hostname,
					       strerror(errno));
					ok = FALSE;
					break;

				/* there's data, see if it's valid */
				} else {
					char           *p = server->secret;

					/* sanity checks on the response data */

					if (total_length < AUTH_HDR_LEN || /* minimum length */
						(ntohs(response->length) != total_length) ||
					        (ntohs(response->length) > RADIUS_PACKET_LEN)) { /* maximum length */
						DPRINT(EVENTLOG_ERROR_TYPE, "RADIUS packet from server %s is corrupted",
						       server->hostname);
						ok = FALSE;
						break;
					}
					/* check that the response ID matches the request ID */
					if (response->id != request->id) {
						DPRINT(EVENTLOG_WARNING_TYPE,
						       "Response packet ID %d doesn't match request packet ID %d: verification of packet fails",
						       response->id,
						       request->id);
						ok = FALSE;
						break;
					}

					/*
					 * Check if we have the data OK
					 */
					if (password) {
						if (old_password) {
							if (!livingston_bug)
								p = old_password;	/* what it should be */
							else
								p = "";	/* what it really is */
						}
					} else {	/* accounting request */
						/*
						 * RFC 2139 p.6 says not do do
						 * this, but the Livingston
						 * 1.16 server disagrees.  If
						 * the user says he wants the
						 * bug, give in.
						 */
						if (livingston_bug)
							p = "";
					}

					if (!verify_packet(p, response, request)) {
						DPRINT(EVENTLOG_ERROR_TYPE,
						       "Packet from RADIUS server %s fails verification: The shared secret is probably incorrect",
						       server->hostname);
						ok = FALSE;
						break;
					}
					/* sanity checks on the response data */
					if (total_length > AUTH_HDR_LEN){
						int	len = ntohs(response->length);
						attribute_t *attr;

						if(len < AUTH_HDR_LEN + 2) { /* We need at least space for one attribute */
							DPRINT(EVENTLOG_ERROR_TYPE,
							"RADIUS packet from server %s is corrupted", server->hostname);
							ok = FALSE;
							break;
						}
						len -= AUTH_HDR_LEN;
						attr = (attribute_t *) & response->data;

						while(attr){
							if (attr->length < 2){
								DPRINT(EVENTLOG_ERROR_TYPE,
							"RADIUS packet from server %s is corrupted", server->hostname);
								ok = FALSE;
								break;
							}
							len -= attr->length;
							if (len == 0) /* We 've reached the end of the packet correctly */
								break;
							if (len < 2) { /* It's either a negative value or less than 2 */
								DPRINT(EVENTLOG_ERROR_TYPE,
							"RADIUS packet from server %s is corrupted", server->hostname);
								ok = FALSE;
								break;
							}
							attr = (attribute_t *) ((char *)attr + attr->length);
						}
					}
					if (ok == FALSE)
						break;

				}

				/*
				 * Whew!  The select is done.  It hasn't timed
				 * out, or errored out.  It's our descriptor.
				 * We've got some data.  It's the right size.
				 * The packet is valid.  NOW, we can skip out
				 * of the select loop, and process the packet.
				 */
				break;
			}
			/* otherwise, we've got data on another descriptor, keep select'ing */
		}

		/* go to the next server if this one didn't respond */
next:
		if (!ok) {
			radius_server_t *old;	/* forget about this server */

			old = server;
			server = server->next;
			conf->server = server;

			_radius_forget(old->secret);
			free(old->hostname);
			free(old);

			if (server) {	/* if there's more servers to check */

				/* get a new authentication vector, and update the passwords */
				get_random_vector(request->vector);
				request->id = request->vector[0];

				/* update passwords, as appropriate */
				if (password) {
					get_random_vector(request->vector);
					if (old_password) {	/* password change request */
						add_password(request, PW_PASSWORD, password, old_password);
						add_password(request, PW_OLD_PASSWORD, old_password, old_password);
					} else {	/* authentication request */
						add_password(request, PW_PASSWORD, password, server->secret);
					}
				}
			}
			continue;

		} else {
			/* we've found one that does respond, forget about the other servers */
			cleanup(server->next);
			server->next = NULL;

			if (!modulePrivate.server)
				modulePrivate.server = radius_dup(server);

			break;
		}
	}

	if (!server) {
		DPRINT(EVENTLOG_ERROR_TYPE, "All RADIUS servers failed to respond");
		return RADIUS_AUTHINFO_UNAVAIL;
	}

	return RADIUS_SUCCESS;
}

/* ---------------------------------------------------------------------------
 * Exported functions
 */

#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { return retval; }

int
authenticate(const char *user, char *password, pGinaInfo *settingsInfo)
{
	int             retval = RADIUS_AUTH_ERR;
	char            recv_buffer[RADIUS_PACKET_LEN];
	char            send_buffer[RADIUS_PACKET_LEN];
	char           *strvalue = NULL;
	AUTH_HDR       *request = (AUTH_HDR *) send_buffer;
	AUTH_HDR       *response = (AUTH_HDR *) recv_buffer;
	radius_conf_t   config;
	int             type;
	int             tries = 3;	/* should we make this configurable? */
	WSADATA         wsa;
	bool            anonymous = false;
	char           *anonymous_filter = NULL;

	modulePrivate.server = NULL;

	debug = regReadBool("debug");
	userlogon = regReadBool("userlogon");
	anonymous = regReadBool("anonymous");
	anonymous_filter = regReadString("anonymous_filter");
	if (anonymous_filter && strlen(anonymous_filter) > MAXNAME) {
		DPRINT(EVENTLOG_ERROR_TYPE, "Anonymous filter too long");
		return RADIUS_FAILURE;
	}

	/* check username */
	if (user) {
		if (strlen(user) > MAXNAME) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Username too long");
			return RADIUS_FAILURE;
		}
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Got username %s", user);
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No username given");
		return RADIUS_FAILURE;
	}

	/* check password */
	if (password) {
		if (strlen(password) > MAXPASS) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Password too long");
			return RADIUS_FAILURE;
		}
		password = strdup(password);
		DPRINT(EVENTLOG_INFORMATION_TYPE,
		       "Got password %s",
		       stars((int)strlen(password)));
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No password given for user %s", user);
		return RADIUS_FAILURE;
	}

	/*
	 * Get the IP address of the authentication server.  Then, open a
	 * socket, and bind it to a port.
	 */
	if (WSAStartup(MAKEWORD(1, 1), &wsa) != 0)
		return RADIUS_AUTHINFO_UNAVAIL;

	retval = initialize(&config, FALSE);
	RADIUS_FAIL_CHECK;

	/* now we've got a socket open, so we've got to clean it up on error */
#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { goto error; }

	/* build and initialize the RADIUS packet */
	request->code = PW_AUTHENTICATION_REQUEST;
	get_random_vector(request->vector);
	request->id = request->vector[0];	/* this should be evenly distributed */

	build_radius_packet(request, user, password, &config);

	/* add a Service-Type (normally Login-User) */
	if (!(type = (int)regReadUlong("service_type")))
		type = PW_LOGIN_USER;
	add_int_attribute(request, PW_USER_SERVICE_TYPE, type);

	if (anonymous){
		if (anonymous_filter){
			if (!strncmp(user,anonymous_filter,strlen(anonymous_filter)))
				add_vendor_int_attribute(request, PW_VENDOR_NTUA, PW_USERLOGON_RESTRICTION,
				PW_USERLOGON_RESTRICTION_ANONYMOUS);
			free(anonymous_filter);
		}
		else{
			add_vendor_int_attribute(request, PW_VENDOR_NTUA, PW_USERLOGON_RESTRICTION,
			PW_USERLOGON_RESTRICTION_ANONYMOUS);
		}
	}

	DPRINT(EVENTLOG_INFORMATION_TYPE, "Sending RADIUS request code %d", request->code);

	retval = talk_radius(&config, request, response, password, NULL, tries);
	RADIUS_FAIL_CHECK;

	DPRINT(EVENTLOG_INFORMATION_TYPE, "Got RADIUS response code %d", response->code);

	/* Whew!  Done the pasword checks, look for an authentication acknowledge. */
	if (response->code == PW_AUTHENTICATION_ACK){
		attribute_t *attr;
		unsigned int value = 0;

		retval = RADIUS_SUCCESS;
		/* Check for a Class attribute. If it exists copy it's contents */
		attr = find_attribute(response, PW_CLASS);
		if ((strvalue = get_string_attribute_value(attr)) != NULL){
			modulePrivate.Class = strdup(strvalue);
		}
		/* Check for a Session-Timeout attribute. If it exists set the sessionTimeout variable */
		attr = find_attribute(response, PW_SESSION_TIMEOUT);
		if (get_int_attribute_value(attr,&value)){
			value = ntohl(value);
			DPRINT(EVENTLOG_INFORMATION_TYPE, "Attribute value: %d", value);
			settingsInfo->sessionTimeout = value;
		}
		if (userlogon){
			vendor_attr_t *vendor;

			DPRINT(EVENTLOG_INFORMATION_TYPE, "Module searching for userlogon attributes");

			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_DRIVENAMES);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->mapPaths = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_HOMEDIR);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->homeDir = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_RESTRICTION);
			if (get_vendor_int_attribute(vendor,&value)){
				value = ntohl(value);
				if (value == PW_USERLOGON_RESTRICTION_ADMIN)
					settingsInfo->isAdmin = true;
				else
					settingsInfo->isAdmin = regReadBool("admin");
			}
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_GROUPNAMES);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->userGroups = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			else
				settingsInfo->userGroups = _tcsdup(TEXT("Users"));
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_USERDESCRIPTION);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->userDescription = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			else
				settingsInfo->userDescription = _tcsdup(TEXT("RADIUS User"));
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_USERFULLNAME);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->userFullName = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_USERPROFILE);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->pathProfile = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
			vendor = find_vendor_attribute( response, PW_VENDOR_NTUA, PW_USERLOGON_USERDOMAIN);
			if ((strvalue = get_vendor_string_attribute_value(vendor)) != NULL){
				settingsInfo->defaultDomain = convert_to_nwchar((const char *)strvalue,
				strlen(strvalue));
			}
		}
		else{
			settingsInfo->isAdmin = regReadBool("admin");
			settingsInfo->userGroups = _tcsdup(TEXT("Users"));
			settingsInfo->userDescription = _tcsdup(TEXT("RADIUS User"));
		}
	}
	else{
		attribute_t *attr;
		char msg[AUTH_STRING_LEN + 30];
		
		memset(msg,0,AUTH_STRING_LEN + 30);
		strcat(msg,"Login incorrect");
		attr = find_attribute(response, PW_REPLY_MESSAGE);
		if ((strvalue = get_string_attribute_value(attr)) != NULL){
			strcat(msg,": ");
			strncat(msg,(const char *)strvalue,strlen(strvalue));
		}
		settingsInfo->errorString = convert_to_nwchar((const char *)msg,strlen(msg));
		retval = RADIUS_AUTH_ERR;	/* authentication failure */

		settingsInfo->isAdmin = regReadBool("admin");
		settingsInfo->userGroups = _tcsdup(TEXT("Users"));
		settingsInfo->userDescription = _tcsdup(TEXT("RADIUS User"));
	}
error:
	DPRINT(EVENTLOG_INFORMATION_TYPE, "Authentication %s",
	       retval == RADIUS_SUCCESS ? "succeeded" : "failed");

	closesocket(config.sockfd);
	WSACleanup();
	_radius_forget(password);
	cleanup(config.server);
	/* keep modulePrivate.server for accounting if we're authenticated */
	if (!((retval == RADIUS_SUCCESS) && regReadBool("accounting")))
		cleanup(modulePrivate.server);
	return retval;
}

#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { return RADIUS_SESSION_ERR; }

int
accounting(const char *user, int status)
{
	int             retval = RADIUS_AUTH_ERR;
	char            recv_buffer[RADIUS_PACKET_LEN];
	char            send_buffer[RADIUS_PACKET_LEN];
	AUTH_HDR       *request = (AUTH_HDR *) send_buffer;
	AUTH_HDR       *response = (AUTH_HDR *) recv_buffer;
	radius_conf_t   config;
	int             type;
	WSADATA         wsa;

	debug = regReadBool("debug");
	userlogon = regReadBool("userlogon");

	/* check username */
	if (user) {
		if (strlen(user) > MAXNAME) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Username too long");
			return RADIUS_FAILURE;
		}
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Got username %s", user);
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No username given");
		return RADIUS_FAILURE;
	}

	/*
	 * Get the IP address of the authentication server.  Then, open a
	 * socket, and bind it to a port.
	 */
	if (WSAStartup(MAKEWORD(1, 1), &wsa) != 0)
		return RADIUS_AUTHINFO_UNAVAIL;

	retval = initialize(&config, TRUE);
	RADIUS_FAIL_CHECK;

	/* now we've got a socket open, so we've got to clean it up on error */
#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { goto error; }

	/* build and initialize the RADIUS packet */
	request->code = PW_ACCOUNTING_REQUEST;
	get_random_vector(request->vector);
	request->id = request->vector[0];	/* this should be evenly distributed */

	build_radius_packet(request, user, NULL, &config);

	add_int_attribute(request, PW_ACCT_STATUS_TYPE, status);

	/* use "service_type_acct", else "service_type", else Login-User */
	if (!(type = (int)regReadUlong("service_type_acct")))
		if (!(type = (int)regReadUlong("service_type")))
			type = PW_LOGIN_USER;
	add_int_attribute(request, PW_USER_SERVICE_TYPE, type);

	if (status == PW_STATUS_START) {
		sgenrand((uint32_t)time(NULL)^getpid());
		sprintf(modulePrivate.s_id, "%08lX", genrand());
	}
	strncpy(recv_buffer, modulePrivate.s_id, 9);

	add_attribute(request, PW_ACCT_SESSION_ID, (unsigned char *)recv_buffer,
	              (int)strlen(recv_buffer));

	add_int_attribute(request, PW_ACCT_AUTHENTIC, PW_AUTH_RADIUS);

	if (status == PW_STATUS_START)
		modulePrivate.login_time = time(NULL);
	else {
		add_int_attribute(request, PW_ACCT_SESSION_TIME, 
			(int)time(NULL) - (int)modulePrivate.login_time);
		add_int_attribute(request, PW_ACCT_TERMINATE_CAUSE, PW_USER_REQUEST);
	}
	
	if (modulePrivate.Class){
		add_attribute(request, PW_CLASS, (unsigned char *)modulePrivate.Class,
			(int)strlen(modulePrivate.Class));
		if (status == PW_STATUS_STOP)
			free(modulePrivate.Class);
	}

	DPRINT(EVENTLOG_INFORMATION_TYPE, "Sending RADIUS request code %d", request->code);

	retval = talk_radius(&config, request, response, NULL, NULL, 1);
	RADIUS_FAIL_CHECK;

	DPRINT(EVENTLOG_INFORMATION_TYPE, "Got RADIUS response code %d", response->code);

	/* Oops!  They don't have the right password.  Complain and die. */
	if (response->code != PW_ACCOUNTING_RESPONSE) {
		retval = RADIUS_PERM_DENIED;
		goto error;
	}
	retval = RADIUS_SUCCESS;
error:
	DPRINT(EVENTLOG_INFORMATION_TYPE, "Accounting %s request %s",
	       status == PW_STATUS_START ? "start" : "stop",
	       retval == RADIUS_SUCCESS ? "succeeded" : "failed");

	if (status == PW_STATUS_STOP)
		cleanup(modulePrivate.server);
	closesocket(config.sockfd);
	WSACleanup();
	cleanup(config.server);
	return retval;
}

/*
 * TODO: Password change not tested yet!
 */
#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { return retval; }

int
chpasswd(const char *user, char *password, char *new_password)
{
	const char     *rhost;
	int             retval = RADIUS_AUTHTOK_ERR;
	radius_conf_t   config;
	char            recv_buffer[RADIUS_PACKET_LEN];
	char            send_buffer[RADIUS_PACKET_LEN];
	AUTH_HDR       *request = (AUTH_HDR *) send_buffer;
	AUTH_HDR       *response = (AUTH_HDR *) recv_buffer;
	attribute_t    *attr;
	WSADATA         wsa;

	debug = regReadBool("debug");

	/* check username */
	if (user) {
		if (strlen(user) > MAXNAME) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Username too long");
			return RADIUS_FAILURE;
		}
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Got username %s", user);
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No username given");
		return RADIUS_FAILURE;
	}

	/* check old password */
	if (password) {
		if (strlen(password) > MAXPASS) {
			DPRINT(EVENTLOG_ERROR_TYPE, "Old password too long");
			return RADIUS_FAILURE;
		}
		password = strdup(password);
		DPRINT(EVENTLOG_INFORMATION_TYPE,
		       "Got old password %s",
		       stars((int)strlen(password)));
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No old password given for user %s", user);
		return RADIUS_FAILURE;
	}

	/* check new password */
	if (new_password) {
		if (strlen(new_password) > MAXPASS) {
			DPRINT(EVENTLOG_ERROR_TYPE, "New password too long");
			return RADIUS_FAILURE;
		}
		new_password = strdup(new_password);
		DPRINT(EVENTLOG_INFORMATION_TYPE,
		       "Got new password %s",
		       stars((int)strlen(new_password)));
	} else {
		DPRINT(EVENTLOG_ERROR_TYPE, "No new password given for user %s", user);
		return RADIUS_FAILURE;
	}

	/*
	 * Get the IP address of the authentication server.  Then, open a
	 * socket, and bind it to a port.
	 */
	if (WSAStartup(MAKEWORD(1, 1), &wsa) != 0)
		return RADIUS_AUTHINFO_UNAVAIL;

	retval = initialize(&config, FALSE);
	RADIUS_FAIL_CHECK;

	/* now we've got a socket open, so we've got to clean it up on error */
#undef RADIUS_FAIL_CHECK
#define RADIUS_FAIL_CHECK if (retval != RADIUS_SUCCESS) { goto error; }

	/* build and initialize the access request RADIUS packet */
	request->code = PW_AUTHENTICATION_REQUEST;
	get_random_vector(request->vector);
	request->id = request->vector[0];	/* this should be evenly distributed */

	build_radius_packet(request, user, password, &config);
	if ((attr = find_attribute(request, PW_USER_SERVICE_TYPE)) != NULL)
		change_int_attribute(attr, PW_AUTHENTICATE_ONLY);
	else
		add_int_attribute(request, PW_USER_SERVICE_TYPE, PW_AUTHENTICATE_ONLY);

	/* tell the server which host the user is coming from */
	if ((rhost = getHostID()) != NULL) {
		DPRINT(EVENTLOG_INFORMATION_TYPE, "Got host IP %s", rhost);
		add_attribute(request, PW_CALLING_STATION_ID, (unsigned char *)rhost,
		              (int)strlen(rhost));
	}
	DPRINT(EVENTLOG_INFORMATION_TYPE, "Sending RADIUS request code %d", request->code);

	retval = talk_radius(&config, request, response, password, NULL, 1);
	RADIUS_FAIL_CHECK;

	DPRINT(EVENTLOG_INFORMATION_TYPE, "Got RADIUS response code %d", response->code);

	/* Oops!  They don't have the right password.  Complain and die. */
	if (response->code != PW_AUTHENTICATION_ACK) {
		_radius_forget(password);
		retval = RADIUS_PERM_DENIED;
		goto error;
	}

	/*
	 * We're now sure it's the right user
	 */

	/* TODO: if (strcmp(password, new_password) == 0) {...} */

	if (!password || !new_password) {	/* ensure we've got passwords */
		retval = RADIUS_AUTHTOK_ERR;
		goto error;
	}

	/* build and initialize the password change request RADIUS packet */
	request->code = PW_PASSWORD_REQUEST;
	get_random_vector(request->vector);
	request->id = request->vector[0];	/* this should be evenly distributed */

	/* the secret here can not be known to the user, so it's the new password */
	_radius_forget(config.server->secret);
	config.server->secret = strdup(password);	/* it's free'd later */

	build_radius_packet(request, user, new_password, &config);
	add_password(request, PW_OLD_PASSWORD, password, password);

	retval = talk_radius(&config, request, response, new_password, password, 1);
	RADIUS_FAIL_CHECK;

	/* Whew!  Done password changing, check for password acknowledge. */
	if (response->code != PW_PASSWORD_ACK) {
		retval = RADIUS_AUTHTOK_ERR;
		goto error;
	}
error:
	DPRINT(EVENTLOG_INFORMATION_TYPE, "Password change %s",
	       retval == RADIUS_SUCCESS ? "succeeded" : "failed");

	closesocket(config.sockfd);
	WSACleanup();
	_radius_forget(password);
	_radius_forget(new_password);
	cleanup(config.server);
	return retval;
}
