//
// $Id: RADIUSplugin.cpp,v 1.6 2005/03/03 18:04:23 hweiss Exp $
//
// RADIUS Plugin for pGina
// Copyright (c) 2004, 2005 Holger Weiss <holger@ZEDAT.FU-Berlin.DE>
// Copyright (c) 2005 Kostas Kalevras <kkalev@noc.ntua.gr>
//
// ---------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc., 59
// Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// ---------------------------------------------------------------------------
//

#include <windows.h>
#include <string.h>

#include "stdafx.h"
#include "RADIUSplugin.h"
#include "RADIUSauth.h"
#include "RADIUSconfig.h"
#include "RADIUShelpers.h"
#include "radius.h"

static bool logged_in = false;
static bool do_accounting = false;

// ---------------------------------------------------------------------------
// RADIUSplugin API

// UserLogin

RADIUSPLUGIN_API BOOL UserLogin(LPTSTR Username, LPTSTR Password, pGinaInfo *settingsInfo)
{
	char *username, *password;
	int   retval;

	// secure our Registry key on every login if "secure_key" is set
	if (regReadBool("secure_key") && !regSecureKey()) {
		settingsInfo->errorString = _tcsdup(TEXT("Can't set access control for registry key"));
		return false;
	}

	// check username length
	if ((Username == NULL) || (_tcslen(Username) == 0)) {
		settingsInfo->errorString = _tcsdup(TEXT("No username given"));
		return false;
	}
	if (_tcslen(Username) > MAXNAME) {
		settingsInfo->errorString = _tcsdup(TEXT("Username too long"));
		return false;
	}

	// don't try to authenticate "administrator" via RADIUS
	if (_wcsicmp(Username, TEXT("administrator")) == 0) {
		settingsInfo->errorString = _tcsdup(TEXT("Administrator not authenticated via RADIUS"));
		return false;
	}

	// check password length
	if (Password == NULL) {
		settingsInfo->errorString = _tcsdup(TEXT("No password given"));
		return false;
	}
	if (_tcslen(Password) > MAXPASS) {
		settingsInfo->errorString = _tcsdup(TEXT("Password too long"));
		return false;
	}

	// if "only_local" is set, check whether user exists on local system
	if (regReadBool("only_local") && !isValidUser(Username)) {
		settingsInfo->errorString = _tcsdup(TEXT("Not a local Windows user"));
		return false;
	}

	// We should convert the RADIUSplugin code to UNICODE.  Tomorrow ;-)
	// In the meantime, we talk UNICODE to pGina, but we convert the
	// UNICODE data we get from pGina to something normal and ignore
	// UNICODE as far as possible.

	username = convert_to_char(Username);
	password = convert_to_char(Password);

	// ask RADIUS server
	retval = authenticate(username, password, settingsInfo);

	// free our username/password from heap
	(void)GlobalFree(username);
	(void)GlobalFree(password);

	// everything went fine, username and password are correct
	if (retval == RADIUS_SUCCESS) {
		settingsInfo->allowPassChange = regReadBool("pw_change");
		logged_in = true;
		return true;
	}

	// user could not be authenticated
	else if (retval == RADIUS_AUTH_ERR);
//		settingsInfo->errorString = _tcsdup(TEXT("Login incorrect"));
	else if (retval == RADIUS_AUTHINFO_UNAVAIL)
		settingsInfo->errorString = _tcsdup(TEXT("Can't connect to RADIUS server"));
	else if (retval == RADIUS_FAILURE)
		settingsInfo->errorString = _tcsdup(TEXT("Fatal error"));
	else
		settingsInfo->errorString = _tcsdup(TEXT("Unknown error"));

	logged_in = false;
	return false;
}

// LoginHook

RADIUSPLUGIN_API void LoginHook(pGinaInfo *settingsInfo)
{
	if ((do_accounting = regReadBool("accounting")) && logged_in) {

		char *username;

		// check username length
		if ((settingsInfo->Username == NULL)
		        || (_tcslen(settingsInfo->Username) == 0)
		        || (_tcslen(settingsInfo->Username) > MAXNAME))
			return;

		username = convert_to_char(settingsInfo->Username);

		(void)accounting(username, PW_STATUS_START);
		(void)GlobalFree(username);
	}
}

// LogoutHook

RADIUSPLUGIN_API void LogoutHook(pGinaInfo *settingsInfo)
{
	if (do_accounting && logged_in) {

		char *username;

		// check username length
		if ((settingsInfo->Username == NULL)
		        || (_tcslen(settingsInfo->Username) == 0)
		        || (_tcslen(settingsInfo->Username) > MAXNAME))
			return;

		username = convert_to_char(settingsInfo->Username);

		(void)accounting(username, PW_STATUS_STOP);
		(void)GlobalFree(username);
	}
	logged_in = false;
}

// ChangeUserPassword (TODO: this is NOT tested yet!)

RADIUSPLUGIN_API BOOL ChangeUserPassword(LPTSTR Username, LPTSTR OldPassword, LPTSTR NewPassword)
{
	if (regReadBool("pw_change")) {

		char *username, *oldpassw, *newpassw;
		int   retval;

		// check username length
		if ((Username == NULL)
		        || (_tcslen(Username) == 0)
		        || (_tcslen(Username) > MAXNAME))
			return false;

		// check old password length
		if ((OldPassword == NULL)
		        || (_tcslen(OldPassword) == 0)
		        || (_tcslen(OldPassword) > MAXPASS))
			return false;

		// check new password length
		if ((NewPassword == NULL)
		        || (_tcslen(NewPassword) == 0)
		        || (_tcslen(NewPassword) > MAXPASS))
			return false;

		username = convert_to_char(Username);
		oldpassw = convert_to_char(OldPassword);
		newpassw = convert_to_char(NewPassword);

		// ask RADIUS server for password change
		retval = chpasswd(username, oldpassw, newpassw);

		// free our username/oldpassw/newpassw from heap
		(void)GlobalFree(username);
		(void)GlobalFree(oldpassw);
		(void)GlobalFree(newpassw);

		if (retval == RADIUS_SUCCESS)
			return true;
	}

	return false;
}

// ChangePluginSettings

RADIUSPLUGIN_API void ChangePluginSettings(void)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	CRADIUSconfig *confDialog = new CRADIUSconfig();
	confDialog->DoModal();
}

// IsRequired

RADIUSPLUGIN_API BOOL IsRequired(void)
{
	return regReadBool("required");
}

// AboutPlugin

RADIUSPLUGIN_API LPTSTR AboutPlugin(void)
{
	return TEXT(VERSION);
}
